from pydantic import BaseModel
from typing import Generic, TypeVar, Optional, Dict, Any, List, Union

from .openapi import danja_openapi

__all__ = [
    "DANJASingleResource",
    "DANJAResource",
    "DANJAResourceList",
    "DANJALink",
    "DANJAError",
    "DANJAErrorList",
    "danja_openapi"
]

ResourceType = TypeVar("ResourceType")


class DANJALink(BaseModel):
    """JSON:API Link"""
    href: str
    rel: Optional[str] = None
    describedby: Optional[str] = None
    title: Optional[str] = None
    type: Optional[str] = None
    hreflang: Optional[str] = None
    meta: Optional[Dict[str, Any]] = None


class DANJASource(BaseModel):
    """JSON:API Source"""
    pointer: str
    parameter: str
    header: str


class DANJAResourceIdentifier(BaseModel):
    """JSON:API Resource Identifier"""
    type: str
    id: str
    lid: Optional[str] = None


class DANJARelationship(BaseModel):
    """JSON:API Relationship"""
    links: Optional[Dict[str, Union[str, DANJALink, None]]] = None
    data: Optional[Dict[str, Union[DANJAResourceIdentifier, List[DANJAResourceIdentifier], None]]] = None  # noqa: E501
    meta: Optional[Dict[str, Any]] = None


class DANJAError(BaseModel):
    """JSON:API Error object"""
    id: Optional[str] = None
    links: Optional[Dict[str, Union[str, DANJALink, None]]] = None
    status: Optional[str] = None
    code: Optional[str] = None
    title: Optional[str] = None
    detail: Optional[str] = None
    source: Optional[Dict[str, DANJASource]] = None
    meta: Optional[Dict[str, Any]] = None


class DANJAErrorList(BaseModel):
    """JSON:API Error list"""
    errors: List[DANJAError]


class DANJASingleResource(BaseModel, Generic[ResourceType]):
    """A single resource. The only JSON:API required field is type"""
    id: Optional[str] = None
    type: str
    lid: Optional[str] = None
    attributes: ResourceType
    relationships: Optional[Dict[str, DANJARelationship]] = None
    links: Optional[Dict[str, Any]] = None
    meta: Optional[Dict[str, Any]] = None


class ResourceResolver():
    @classmethod
    def resolve_resource_name(cls, resource) -> str:
        return str(resource.model_config.get(
            "resource_name",
            resource.__class__.__name__.lower()
        ))

    @classmethod
    def resolve_resource_id(cls, resource) -> Optional[str]:
        for field_name, field in resource.model_fields.items():
            if isinstance(field.json_schema_extra, dict):
                if "resource_id" in field.json_schema_extra:
                    return field_name
            elif isinstance(field.schema_extra, dict):  # Support for older SQLModel
                if "resource_id" in field.json_schema_extra:
                    return field_name

        return None


class DANJAResource(BaseModel, ResourceResolver, Generic[ResourceType]):
    """JSON:API base for a single resource"""
    data: DANJASingleResource[ResourceType]
    links: Optional[Dict[str, Union[str, DANJALink, None]]] = None
    meta: Optional[Dict[str, Any]] = None
    included: Optional[List[Dict[str, Any]]] = None

    @property
    def resource(self) -> ResourceType:
        return self.data.attributes

    @classmethod
    def from_basemodel(
            cls,
            resource: ResourceType,
            resource_name: Optional[str] = None,
            resource_id: Optional[str] = None
    ) -> "DANJAResource":
        try:
            if not resource_name:
                """
                No resource name supplied, look for one in the model config
                or failing that use the resource class name
                """
                resource_name = cls.resolve_resource_name(resource)

            if not resource_id:
                """
                No resource ID fields supplied, look for one in the model config
                or failing that if there's
                """
                resource_id = cls.resolve_resource_id(resource)
                if not resource_id:
                    raise Exception(f"No fields defined in {resource_name}")

            values = {
                "type": resource_name,
                "lid": None,
                "attributes": resource
            }

            id_value = object.__getattribute__(resource, resource_id)
            if id_value:
                values["id"] = str(id_value)

            return cls(data=DANJASingleResource(**values))
        except AttributeError:
            raise Exception(
                f"Resource ID field not found in {resource_name}: {resource_id}"
            )


class DANJAResourceList(BaseModel, ResourceResolver, Generic[ResourceType]):
    """JSON:API base for a list of resources"""
    data: List[DANJASingleResource[ResourceType]]
    links: Optional[Dict[str, Union[str, DANJALink, None]]] = None
    meta: Optional[Dict[str, Any]] = None
    included: Optional[List[DANJASingleResource]] = None

    @property
    def resources(self) -> List[ResourceType]:
        return [data.attributes for data in self.data]

    @classmethod
    def from_basemodel_list(
            cls,
            resources: List[ResourceType],
            resource_name: Optional[str] = None,
            resource_id: Optional[str] = None
    ) -> "DANJAResourceList":
        try:
            if len(resources) > 0:
                resource = resources[0]

                if not resource_name:
                    """
                    No resource name supplied, look for one in the model config
                    or failing that use the resource class name
                    """
                    resource_name = cls.resolve_resource_name(resource)

                if not resource_id:
                    """
                    No resource ID fields supplied, look for one in the model config
                    or failing that if there's
                    """
                    resource_id = cls.resolve_resource_id(resource)
                    if not resource_id:
                        raise Exception(f"No fields defined in {resource_name}")

            data: List[DANJASingleResource] = []
            for sub_resource in resources:
                values = {
                    "type": resource_name,
                    "lid": None,
                    "attributes": sub_resource
                }
                id_value = object.__getattribute__(sub_resource, resource_id or "")
                if id_value:
                    values["id"] = str(id_value)
                data.append(DANJASingleResource(**values))

            return cls(data=data)
        except AttributeError:
            raise Exception(
                f"Resource ID field not found in {resource_name}: {resource_id}"
            )



from typing import Dict, Any
import dpath


def danja_openapi(openapi_schema: Dict[str, Any]) -> Dict[str, Any]:
    """
    This is an optional function that will assist in de-cluttering the OpenAPI
    schema space. This function will de-reference and flatten a lot of the
    DANJA namespace while leaving non-DANJA classes alone.
    """
    
    # Get all referenced schemas
    ref_schemas = set([
        result[1][21:]
        for result in dpath.search(openapi_schema, "paths/**/$ref", yielded=True)
    ])

    # Resolve all schema references
    ex_schemas = {}
    for key, value in openapi_schema.get("components", {}).get("schemas", {}).items():
        # Flatten and add to the schemas
        for result in dpath.search(value, "/**/$ref", yielded=True):
            parent, _, _ = result[0].rpartition("/")
            dpath.set(value, parent, dpath.get(openapi_schema, result[1][1:]))
        
        if key in ref_schemas:
            ex_schemas[key] = value

    # Rename the DANJA objects
    # DANJAErrorList -> Error_List
    # DANJAResourceList_model -> model_List
    # DANJAResource_model -> model
    rn_schemas = {}
    for key in ex_schemas:
        old_key = f"#/components/schemas/{key}"
        # Find the key in the paths and rename the key and reference
        if key.startswith("DANJAResource_"):
            new_key = key.replace("DANJAResource_", "")
            # Find all path references to the old key
            for result in dpath.search(
                openapi_schema,
                "paths|**|$ref",
                yielded=True,
                separator="|"
            ):
                if result[1] == old_key:
                    # Rename to the new key
                    dpath.set(
                        openapi_schema,
                        result[0],
                        f"#/components/schemas/{new_key[:-1]}",
                        separator="|"
                    )

            rn_schemas[new_key[:-1]] = ex_schemas[key]
        elif key.startswith("DANJAResourceList_"):
            new_key = key.replace("DANJAResourceList_", "")
            # Find all path references to the old key
            for result in dpath.search(
                openapi_schema,
                "paths|**|$ref",
                yielded=True,
                separator="|"
            ):
                if result[1] == old_key:
                    # Rename to the new key
                    dpath.set(
                        openapi_schema,
                        result[0],
                        f"#/components/schemas/{new_key}List",
                        separator="|"
                    )

            rn_schemas[f"{new_key}List"] = ex_schemas[key]
        elif key == "DANJAErrorList":
            new_key = "Error_List"
            # Find all path references to the old key
            for result in dpath.search(
                openapi_schema,
                "paths|**|$ref",
                yielded=True,
                separator="|"
            ):
                if result[1] == old_key:
                    # Rename to the new key
                    dpath.set(
                        openapi_schema,
                        result[0],
                        f"#/components/schemas/{new_key}",
                        separator="|"
                    )

            rn_schemas[new_key] = ex_schemas[key]
        else:
            # Don't rename
            rn_schemas[key] = ex_schemas[key]

    openapi_schema["components"]["schemas"] = rn_schemas

    return openapi_schema



import pytest
import json
from typing import Optional
from pathlib import Path
from pydanja import DANJAResource, DANJAResourceList
from pydantic import BaseModel, Field


class FixtureTestType(BaseModel):
    """
    A test type for Pydantic and pydanja
    """
    # If we use ID, then we must alias it to avoid clashes with Python
    fixture_testtype_id: Optional[int] = Field(
        alias="id",
        default=None,
        json_schema_extra={
            "resource_id": True
        }
    )
    name: str
    description: str


@pytest.fixture(scope="function")
def resource(request):
    with open(Path("tests", "fixtures", request.param), "r") as json_fixture:
        return json.load(json_fixture)


@pytest.mark.parametrize("resource", ["single_resource_with_id.json"], indirect=True)
def test_it_creates_a_container_from_base_model_with_id(resource):
    """
    Test that we can create a JSON:API container for a single resource
    from a pydantic BaseModel instance
    """
    # BaseModel including ID
    basemodel_instance = FixtureTestType(
        id=1,
        name="Stuff!",
        description="This is desc!"
    )

    # Create the JSON:API container
    new_resource = DANJAResource.from_basemodel(basemodel_instance)

    schema = new_resource.model_json_schema()

    # Check schema
    assert(schema == resource)

    # Check resource data
    assert(new_resource.resource == basemodel_instance)


@pytest.mark.parametrize("resource", ["single_resource_without_id.json"], indirect=True)
def test_it_creates_a_container_from_base_model_without_id(resource):
    """
    Test that we can create a JSON:API container for a single resource
    from a pydantic BaseModel instance
    """
    # BaseModel without ID
    basemodel_instance = FixtureTestType(
        name="Stuff!",
        description="This is desc!"
    )

    # Create the JSON:API container
    new_resource = DANJAResource.from_basemodel(basemodel_instance)

    schema = new_resource.model_json_schema()

    # Check schema
    assert(schema == resource)

    # Check resource data
    assert(new_resource.resource == basemodel_instance)


@pytest.mark.parametrize("resource", ["multiple_resource_with_id.json"], indirect=True)
def test_it_creates_a_container_list_from_base_model_with_id(resource):
    """
    Test that we can create a JSON:API container for a single resource
    from a pydantic BaseModel instance
    """
    # BaseModel including ID
    basemodel_instances = [
        FixtureTestType(
            id=1,
            name="Stuff!",
            description="This is desc!"
        ),
        FixtureTestType(
            id=2,
            name="More Stuff!",
            description="This is more desc!"
        )
    ]

    # Create the JSON:API container
    new_resources = DANJAResourceList.from_basemodel_list(basemodel_instances)

    schema = new_resources.model_json_schema()

    # Check schema
    assert(schema == resource)

    # Check resource data
    assert(len(new_resources.resources) == len(basemodel_instances))


@pytest.mark.parametrize("resource", ["multiple_resource_without_id.json"], indirect=True)
def test_it_creates_a_container_list_from_base_model_without_id(resource):
    """
    Test that we can create a JSON:API container for a single resource
    from a pydantic BaseModel instance
    """
    # BaseModel including ID
    basemodel_instances = [
        FixtureTestType(
            name="Stuff!",
            description="This is desc!"
        ),
        FixtureTestType(
            name="More Stuff!",
            description="This is more desc!"
        )
    ]

    # Create the JSON:API container
    new_resources = DANJAResourceList.from_basemodel_list(basemodel_instances)

    schema = new_resources.model_json_schema()

    # Check schema
    assert(schema == resource)

    # Check resource data
    assert(len(new_resources.resources) == len(basemodel_instances))
